#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Generator, Tuple, Any

import oser.accelerator as oser

from oser.types import OserElementType
from htf.modbus.core import (
    FunctionCode as _FunctionCode,
    ModbusPDUBase as _ModbusPDUBase,
    ModbusCrcBase,
    ModbusLRCBase,
)


class ModbusCrc(ModbusCrcBase):
    """
    Calculates a reversed IBM-CRC-16 (0x8005) -> 0xa001 to be used with Modbus
    """

    def __init__(self) -> None:
        """"""
        super(ModbusCrc, self).__init__()

    def crc(self, data: bytes) -> int:
        return super(ModbusCrc, self).crc(data=data)

    def encode(self, full_data: bytes = b"", context_data: bytes = b"") -> bytes:
        """
        Return the encoded binary string.

        Args:
            full_data: the binary data string until the part to be encoded.
                The user normally does not need to supply this.
            context_data: the binary data of the current context.
                The user normally does not need to supply this.
        """
        return super(ModbusCrc, self).encode(full_data=full_data, context_data=context_data)

    def decode(self, data: bytes, full_data: bytes = b"", context_data: bytes = b"") -> int:
        """
        Decode a binary string and return the number
        of bytes that were decoded.

        Args:
            data: the data buffer that is decoded.
            full_data: the binary data string until the part to be decoded.
                The user normally does not need to supply this.
            context_data: the binary data of the current context.
                The user normally does not need to supply this.

        Returns:
            the number of bytes that were decoded.
        """
        return super(ModbusCrc, self).decode(data=data, full_data=full_data, context_data=context_data)


class ModbusLRC(ModbusLRCBase):
    """
    Compute the logitudinal redudancy check for Modbus.
    """

    def lrc(self, data: bytes) -> int:
        return super(ModbusLRC, self).lrc(data)

    def encode(self, full_data: bytes = b"", context_data: bytes = b"") -> bytes:
        """
        Return the encoded binary string.

        Args:
            full_data: the binary data string until the part to be encoded.
                The user normally does not need to supply this.
            context_data: the binary data of the current context.
                The user normally does not need to supply this.
        """
        return super(ModbusLRC, self).encode(full_data=full_data, context_data=context_data)

    def decode(self, data: bytes, full_data: bytes = b"", context_data: bytes = b"") -> int:
        """
        Decode a binary string and return the number
        of bytes that were decoded.

        Args:
            data: the data buffer that is decoded.
            full_data: the binary data string until the part to be decoded.
                The user normally does not need to supply this.
            context_data: the binary data of the current context.
                The user normally does not need to supply this.

        Returns:
            the number of bytes that were decoded.
        """
        return super(ModbusLRC, self).decode(data=data, full_data=full_data, context_data=context_data)


def _request_generator() -> Generator[Tuple[int, OserElementType], None, None]:
    yield 1, oser.Lazy(ReadCoilsRequest)
    yield 2, oser.Lazy(ReadDiscreteInputsRequest)
    yield 3, oser.Lazy(ReadHoldingRegistersRequest)
    yield 4, oser.Lazy(ReadInputRegistersRequest)
    yield 5, oser.Lazy(WriteSingleCoilRequest)
    yield 6, oser.Lazy(WriteSingleRegisterRequest)
    yield 7, oser.Lazy(ReadExceptionStatusRequest)
    yield 8, oser.Lazy(DiagnosticsRequest)
    yield 11, oser.Lazy(GetCommEventCounterRequest)
    yield 12, oser.Lazy(GetCommEventLogRequest)
    yield 15, oser.Lazy(WriteMultipleCoilsRequest)
    yield 16, oser.Lazy(WriteMultipleRegistersRequest)
    yield 17, oser.Lazy(ReportServerIdRequest)
    yield 20, oser.Lazy(ReadExtendedMemoryRecordRequest)
    yield 21, oser.Lazy(WriteExtendedMemoryRecordRequest)
    yield 22, oser.Lazy(MaskWriteRequest)
    yield 23, oser.Lazy(ReadWriteMultipleRegistersRequest)
    yield 24, oser.Lazy(ReadFifoQueueRequest)
    yield 43, oser.Lazy(EncapsulatedInterfaceTransportRequest)


def _response_generator() -> Generator[Tuple[int, OserElementType], None, None]:
    yield 1, oser.Lazy(ReadCoilsResponse)
    yield 2, oser.Lazy(ReadDiscreteInputsResponse)
    yield 3, oser.Lazy(ReadHoldingRegistersResponse)
    yield 4, oser.Lazy(ReadInputRegistersResponse)
    yield 5, oser.Lazy(WriteSingleCoilResponse)
    yield 6, oser.Lazy(WriteSingleRegisterResponse)
    yield 7, oser.Lazy(ReadExceptionStatusResponse)
    yield 8, oser.Lazy(DiagnosticsResponse)
    yield 11, oser.Lazy(GetCommEventCounterResponse)
    yield 12, oser.Lazy(GetCommEventLogResponse)
    yield 15, oser.Lazy(WriteMultipleCoilsResponse)
    yield 16, oser.Lazy(WriteMultipleRegistersResponse)
    yield 17, oser.Lazy(ReportServerIdResponse)
    yield 20, oser.Lazy(ReadExtendedMemoryRecordResponse)
    yield 21, oser.Lazy(WriteExtendedMemoryRecordResponse)
    yield 22, oser.Lazy(MaskWriteResponse)
    yield 23, oser.Lazy(ReadWriteMultipleRegistersResponse)
    yield 24, oser.Lazy(ReadFifoQueueResponse)
    yield 43, oser.Lazy(EncapsulatedInterfaceTransportResponse)


class FunctionCode(_FunctionCode):
    def is_exception(self) -> bool:
        """
        Check if the current function code is an exception (msb is set).

        Returns:
            ``True`` if the current function code is an exception. Else ``False``.
        """
        return super(FunctionCode, self).is_exception()  # pragma: no cover


class ModbusPDUBase(_ModbusPDUBase):
    """
    ModbusPDUBase is the Modbus PDU base-class.
    """

    def __init__(self) -> None:
        super(ModbusPDUBase, self).__init__()


class RequestPDU(ModbusPDUBase):
    """
    RequestPDU is the request protocol data unit class.
    """

    def __init__(self) -> None:
        super(RequestPDU, self).__init__()
        self.data = oser.Switch(lambda ctx: ctx.function_code.get(), values={}, default=ExceptionResponse())

        self._update_requests()

    def _update_requests(self) -> None:
        for key, value in self.requests_generator():
            self.data.set_value(key, value)

    def requests_generator(self) -> Generator[Tuple[int, OserElementType], None, None]:
        """
        Yields function code, abstraction tuples.
        """
        for key, value in _request_generator():
            yield key, value


class ResponsePDU(ModbusPDUBase):
    """
    ResponsePDU is the response protocol data unit class.
    """

    def __init__(self) -> None:
        super(ResponsePDU, self).__init__()
        self.data = oser.Switch(
            lambda ctx: ctx.function_code.get() & ~0x80,
            values={},
            default=ExceptionResponse(),
        )

        self._update_responses()

    def _update_responses(self) -> None:
        for key, value in self.responses_generator():
            self.data.set_value(key, value)

    def responses_generator(self) -> Generator[Tuple[int, OserElementType], None, None]:
        """
        Yields function code, abstraction tuples.
        """
        for key, value in _response_generator():
            yield key, value


class ExceptionResponse(oser.ByteStruct):
    """
    ExceptionResponse is the response class representing invalid requests
    or device failures. It's the base-class of all responses.
    """

    def __init__(self) -> None:
        super(ExceptionResponse, self).__init__()
        self.exception_code = oser.If(
            lambda ctx: ctx.up().function_code.is_exception(),
            oser.Enum(
                prototype=oser.UBInt8,
                values={
                    "illegal_function": 0x01,
                    "illegal_data_address": 0x02,
                    "illegal_data_value": 0x03,
                    "server_device_failure": 0x04,
                    "acknowledge": 0x05,
                    "server_device_busy": 0x06,
                    "negative_acknowledge": 0x07,
                    "memory_parity_error": 0x08,
                    "gateway_path_unavailable": 0x0A,
                    "gateway_target_device_failed_to_respond": 0x0B,
                },
            ),
        )

    def _items(self) -> Generator[Tuple[str, OserElementType], None, None]:
        for name, value in super(ExceptionResponse, self)._items():
            yield name, value
            is_error = self.up().function_code.is_exception()
            if is_error is True and name == "exception_code":
                return


class ReadCoilsRequest(oser.ByteStruct):
    """
    Representation of a read coils request.
    """

    def __init__(self) -> None:
        super(ReadCoilsRequest, self).__init__()
        self.starting_address = oser.UBInt16()
        self.quantity_of_coils = oser.UBInt16()


class ReadCoilsResponse(ExceptionResponse):
    """
    Representation of a read coils response.
    """

    def __init__(self) -> None:
        super(ReadCoilsResponse, self).__init__()
        self.byte_count = oser.UBInt8()
        self.coil_status = oser.Array(length=lambda ctx: ctx.byte_count.get(), prototype=oser.UBInt8)


class ReadDiscreteInputsRequest(oser.ByteStruct):
    """
    Representation of a read discrete inputs request.
    """

    def __init__(self) -> None:
        super(ReadDiscreteInputsRequest, self).__init__()
        self.starting_address = oser.UBInt16()
        self.quantity_of_inputs = oser.UBInt16()


class ReadDiscreteInputsResponse(ExceptionResponse):
    """
    Representation of a read discrete inputs response.
    """

    def __init__(self) -> None:
        super(ReadDiscreteInputsResponse, self).__init__()
        self.byte_count = oser.UBInt8()
        self.input_status = oser.Array(length=lambda ctx: ctx.byte_count.get(), prototype=oser.UBInt8)


class ReadHoldingRegistersRequest(oser.ByteStruct):
    """
    Representation of a read holding registers request.
    """

    def __init__(self) -> None:
        super(ReadHoldingRegistersRequest, self).__init__()
        self.starting_address = oser.UBInt16()
        self.quantity_of_registers = oser.UBInt16()


class ReadHoldingRegistersResponse(ExceptionResponse):
    """
    Representation of a read holding registers response.
    """

    def __init__(self) -> None:
        super(ReadHoldingRegistersResponse, self).__init__()
        self.byte_count = oser.UBInt8()
        self.register_value = oser.Array(length=lambda ctx: ctx.byte_count.get() // 2, prototype=oser.UBInt16)


class ReadInputRegistersRequest(oser.ByteStruct):
    """
    Representation of a read input registers request.
    """

    def __init__(self) -> None:
        super(ReadInputRegistersRequest, self).__init__()
        self.starting_address = oser.UBInt16()
        self.quantity_of_registers = oser.UBInt16()


class ReadInputRegistersResponse(ExceptionResponse):
    """
    Representation of a read input registers response.
    """

    def __init__(self) -> None:
        super(ReadInputRegistersResponse, self).__init__()
        self.byte_count = oser.UBInt8()
        self.input_registers = oser.Array(length=lambda ctx: ctx.byte_count.get() // 2, prototype=oser.UBInt16)


class WriteSingleCoilRequest(oser.ByteStruct):
    """
    Representation of a write single coil request.
    """

    def __init__(self) -> None:
        super(WriteSingleCoilRequest, self).__init__()
        self.output_address = oser.UBInt16()
        self.output_value = oser.UBInt16()


class WriteSingleCoilResponse(ExceptionResponse):
    """
    Representation of a write single coil response.
    """

    def __init__(self) -> None:
        super(WriteSingleCoilResponse, self).__init__()
        self.output_address = oser.UBInt16()
        self.output_value = oser.UBInt16()


class WriteSingleRegisterRequest(oser.ByteStruct):
    """
    Representation of a write single register request.
    """

    def __init__(self) -> None:
        super(WriteSingleRegisterRequest, self).__init__()
        self.register_address = oser.UBInt16()
        self.register_value = oser.UBInt16()


class WriteSingleRegisterResponse(ExceptionResponse):
    """
    Representation of a write single register response.
    """

    def __init__(self) -> None:
        super(WriteSingleRegisterResponse, self).__init__()
        self.register_address = oser.UBInt16()
        self.register_value = oser.UBInt16()


class ReadExceptionStatusRequest(oser.ByteStruct):
    """
    Representation of a read exception status request.
    """

    pass


class ReadExceptionStatusResponse(ExceptionResponse):
    """
    Representation of a read exception status response.
    """

    def __init__(self) -> None:
        super(ReadExceptionStatusResponse, self).__init__()
        self.output_data = oser.UBInt8()


sub_function_codes = {
    "return_query_data": 0x00,
    "restart_communications_option": 0x01,
    "return_diagnostics_register": 0x02,
    "change_ascii_input_delimiter": 0x03,
    "force_listen_only_mode": 0x04,
    "clear_counters_and_diagnostic_register": 0x0A,
    "return_bus_message_count": 0x0B,
    "return_bus_communication_error_count": 0x0C,
    "return_bus_exception_error_count": 0x0D,
    "return_server_message_count": 0x0E,
    "return_server_no_response_count": 0x0F,
    "return_server_nak_count": 0x10,
    "return_server_busy_count": 0x11,
    "return_bus_character_overrun_count": 0x12,
    "clear_overrun_counter_and_flag": 0x14,
}


class VariableLengthData(oser.String):
    """
    VariableLengthData is a class that holds an arbitrary amount of data.
    Its length depends on the protocol it's embedded in.
    """

    def _is_serial_rtu(self) -> bool:
        return self.root().__class__.__name__.startswith("SerialRTUModbusRe")  # type: ignore

    def _is_serial_ascii(self) -> bool:
        return self.root().__class__.__name__.startswith("SerialASCIIModbusRe")  # type: ignore

    def get(self) -> bytes:
        """
        Return the value.
        """
        return self._value  # type: ignore

    def encode(self, full_data: bytes = b"", context_data: bytes = b"") -> bytes:
        """
        Return the encoded binary string.

        Args:
            full_data: the binary data string until the part to be encoded.
                The user normally does not need to supply this.
            context_data: the binary data of the current context.
                The user normally does not need to supply this.
        """
        self.set_length(len(self.get()))
        return super(VariableLengthData, self).encode(full_data=full_data, context_data=context_data)

    def decode(self, data: bytes, full_data: bytes = b"", context_data: bytes = b"") -> int:
        """
        Decode a binary string and return the number
        of bytes that were decoded.

        Args:
            data: the data buffer that is decoded.
            full_data: the binary data string until the part to be decoded.
                The user normally does not need to supply this.
            context_data: the binary data of the current context.
                The user normally does not need to supply this.

        Returns:
            the number of bytes that were decoded.
        """
        self.set_length(len(data))

        if self._is_serial_rtu():
            self.set_length(len(data) - ModbusCrc().get_byte_size())

        if self._is_serial_ascii():
            self.set_length(len(data) - ModbusLRC().get_byte_size())

        return super(VariableLengthData, self).decode(data, full_data=full_data, context_data=context_data)


class DiagnosticsRequest(oser.ByteStruct):
    """
    Representation of a diagnostics request.
    """

    def __init__(self) -> None:
        super(DiagnosticsRequest, self).__init__()
        self.sub_function = oser.Enum(prototype=oser.UBInt16, values=sub_function_codes)
        self.data = VariableLengthData(
            length=lambda ctx: 2 if self.sub_function.get_value() != 0 else 0,
            value=b"\x00\x00",
        )  # pragma: no cover


class DiagnosticsResponse(ExceptionResponse):
    """
    Representation of a diagnostics response.
    """

    def __init__(self) -> None:
        super(DiagnosticsResponse, self).__init__()
        self.sub_function = oser.Enum(prototype=oser.UBInt16, values=sub_function_codes)
        self.data = VariableLengthData()


class GetCommEventCounterRequest(oser.ByteStruct):
    """
    Representation of a get communication event counter request.
    """

    pass


class GetCommEventCounterResponse(ExceptionResponse):
    """
    Representation of a get communication event counter response.
    """

    def __init__(self) -> None:
        super(GetCommEventCounterResponse, self).__init__()
        self.status = oser.UBInt16()
        self.event_count = oser.UBInt16()


class GetCommEventLogRequest(oser.ByteStruct):
    """
    Representation of a get communication event log request.
    """

    pass


class GetCommEventLogResponse(ExceptionResponse):
    """
    Representation of a get communication event log response.
    """

    def __init__(self) -> None:
        super(GetCommEventLogResponse, self).__init__()
        self.byte_count = oser.UBInt8()
        self.status = oser.UBInt16()
        self.event_count = oser.UBInt16()
        self.message_count = oser.UBInt16()
        self.events = oser.Array(length=lambda ctx: ctx.byte_count.get() - 6, prototype=oser.UBInt8)


class WriteMultipleCoilsRequest(oser.ByteStruct):
    """
    Representation of a write multiple coils request.
    """

    def __init__(self) -> None:
        super(WriteMultipleCoilsRequest, self).__init__()
        self.starting_address = oser.UBInt16()
        self.quantity_of_outputs = oser.UBInt16()
        self.byte_count = oser.UBInt8()
        self.outputs_value = oser.Array(length=lambda ctx: ctx.byte_count.get(), prototype=oser.UBInt8)


class WriteMultipleCoilsResponse(ExceptionResponse):
    """
    Representation of a write multiple coils response.
    """

    def __init__(self) -> None:
        super(WriteMultipleCoilsResponse, self).__init__()
        self.starting_address = oser.UBInt16()
        self.quantity_of_outputs = oser.UBInt16()


class WriteMultipleRegistersRequest(oser.ByteStruct):
    """
    Representation of a write multiple registers request.
    """

    def __init__(self) -> None:
        super(WriteMultipleRegistersRequest, self).__init__()
        self.starting_address = oser.UBInt16()
        self.quantity_of_registers = oser.UBInt16()
        self.byte_count = oser.UBInt8()
        self.outputs_value = oser.Array(length=lambda ctx: ctx.byte_count.get() // 2, prototype=oser.UBInt16)


class WriteMultipleRegistersResponse(ExceptionResponse):
    """
    Representation of a write multiple registers response.
    """

    def __init__(self) -> None:
        super(WriteMultipleRegistersResponse, self).__init__()
        self.starting_address = oser.UBInt16()
        self.quantity_of_registers = oser.UBInt16()


class ReportServerIdRequest(oser.ByteStruct):
    """
    Representation of a report server id request.
    """

    pass


class ReportServerIdResponse(ExceptionResponse):
    """
    Representation of a report server id response.
    """

    def __init__(self) -> None:
        super(ReportServerIdResponse, self).__init__()
        self.byte_count = oser.UBInt8(8)
        self.id = oser.UBInt8(0)  # Micro 84
        self.run_indicator_status = oser.UBInt8(0xFF)
        self.data = oser.Data(length=lambda ctx: ctx.byte_count.get() - 2)


class ReadExtendedMemoryRecordSubRequest(oser.ByteStruct):
    """
    Representation of a read extended memory record sub request.
    """

    def __init__(self) -> None:
        super(ReadExtendedMemoryRecordSubRequest, self).__init__()
        self.reference_type = oser.UBInt8()
        self.memory_number = oser.UBInt16()
        self.record_number = oser.UBInt16()
        self.record_length = oser.UBInt16()


class ReadExtendedMemoryRecordSubResponse(oser.ByteStruct):
    """
    Representation of a read extended memory record sub request.
    """

    def __init__(self) -> None:
        super(ReadExtendedMemoryRecordSubResponse, self).__init__()
        self.response_length = oser.UBInt8()
        self.reference_type = oser.UBInt8()
        self.record_data = oser.Array(length=lambda ctx: ctx.response_length.get() // 2, prototype=oser.UBInt16)


def _get_checksum_length(self: Any) -> int:
    checksum_length = 0
    root_class_name = self.root().__class__.__name__

    if root_class_name.startswith("SerialRTUModbusRe"):
        checksum_length = ModbusCrc().get_byte_size()

    if root_class_name.startswith("SerialASCIIModbusRe"):
        checksum_length = ModbusLRC().get_byte_size()

    return checksum_length


class ReadExtendedMemoryRecordRequest(oser.ByteStruct):
    """
    Representation of a read extended memory record request.
    """

    def __init__(self) -> None:
        super(ReadExtendedMemoryRecordRequest, self).__init__()
        self.byte_count = oser.UBInt8()

        self.sub_requests = oser.RepeatUntil(
            length=1,
            stop_condition=lambda ctx, data, full_data, context_data: len(data) <= _get_checksum_length(ctx),
            prototype=ReadExtendedMemoryRecordSubRequest,
        )

    def encode(self, full_data: bytes = b"", context_data: bytes = b"") -> bytes:
        self.byte_count.set(self.sub_requests.get_byte_size())
        return super(ReadExtendedMemoryRecordRequest, self).encode(full_data=full_data, context_data=context_data)


class ReadExtendedMemoryRecordResponse(ExceptionResponse):
    """
    Representation of a read extended memory record response.
    """

    def __init__(self) -> None:
        super(ReadExtendedMemoryRecordResponse, self).__init__()
        self.byte_count = oser.UBInt8()

        self.sub_responses = oser.RepeatUntil(
            length=1,
            stop_condition=lambda ctx, data, full_data, context_data: len(data) <= _get_checksum_length(ctx),
            prototype=ReadExtendedMemoryRecordSubResponse,
        )

    def encode(self, full_data: bytes = b"", context_data: bytes = b"") -> bytes:
        self.byte_count.set(self.sub_responses.get_byte_size())
        return super(ReadExtendedMemoryRecordResponse, self).encode(full_data=full_data, context_data=context_data)


class WriteExtendedMemoryRecordSubRequest(oser.ByteStruct):
    """
    Representation of a write extended memory record sub request.
    """

    def __init__(self) -> None:
        super(WriteExtendedMemoryRecordSubRequest, self).__init__()
        self.reference_type = oser.UBInt8()
        self.memory_number = oser.UBInt16()
        self.record_number = oser.UBInt16()
        self.record_length = oser.UBInt16()
        self.record_data = oser.Array(length=lambda ctx: ctx.record_length.get(), prototype=oser.UBInt16)


class WriteExtendedMemoryRecordRequest(oser.ByteStruct):
    """
    Representation of a write extended memory record request.
    """

    def __init__(self) -> None:
        super(WriteExtendedMemoryRecordRequest, self).__init__()
        self.byte_count = oser.UBInt8()

        self.sub_requests = oser.RepeatUntil(
            length=1,
            stop_condition=lambda ctx, data, full_data, context_data: len(data) <= _get_checksum_length(ctx),
            prototype=WriteExtendedMemoryRecordSubRequest,
        )

    def encode(self, full_data: bytes = b"", context_data: bytes = b"") -> bytes:
        self.byte_count.set(self.sub_requests.get_byte_size())
        return super(WriteExtendedMemoryRecordRequest, self).encode(full_data=full_data, context_data=context_data)


class WriteExtendedMemoryRecordResponse(ExceptionResponse):
    """
    Representation of a write extended memory record response.
    """

    def __init__(self) -> None:
        super(WriteExtendedMemoryRecordResponse, self).__init__()
        self.byte_count = oser.UBInt8()

        self.sub_responses = oser.RepeatUntil(
            length=1,
            stop_condition=lambda ctx, data, full_data, context_data: len(data) <= _get_checksum_length(ctx),
            prototype=WriteExtendedMemoryRecordSubRequest,
        )

    def encode(self, full_data: bytes = b"", context_data: bytes = b"") -> bytes:
        self.byte_count.set(self.sub_responses.get_byte_size())
        return super(WriteExtendedMemoryRecordResponse, self).encode(full_data=full_data, context_data=context_data)


class MaskWriteRequest(oser.ByteStruct):
    """
    Representation of a mask write request.
    """

    def __init__(self) -> None:
        super(MaskWriteRequest, self).__init__()
        self.reference_address = oser.UBInt16()
        self.and_mask = oser.UBInt16()
        self.or_mask = oser.UBInt16()


class MaskWriteResponse(ExceptionResponse):
    """
    Representation of a mask write response.
    """

    def __init__(self) -> None:
        super(MaskWriteResponse, self).__init__()
        self.reference_address = oser.UBInt16()
        self.and_mask = oser.UBInt16()
        self.or_mask = oser.UBInt16()


class ReadWriteMultipleRegistersRequest(oser.ByteStruct):
    """
    Representation of a write multiple registers request.
    """

    def __init__(self) -> None:
        super(ReadWriteMultipleRegistersRequest, self).__init__()
        self.read_starting_address = oser.UBInt16()
        self.quantity_to_read = oser.UBInt16()
        self.write_starting_address = oser.UBInt16()
        self.quantity_to_write = oser.UBInt16()
        self.write_byte_count = oser.UBInt8()
        self.write_registers_value = oser.Array(
            length=lambda ctx: ctx.write_byte_count.get() // 2, prototype=oser.UBInt16
        )


class ReadWriteMultipleRegistersResponse(ExceptionResponse):
    """
    Representation of a write multiple registers response.
    """

    def __init__(self) -> None:
        super(ReadWriteMultipleRegistersResponse, self).__init__()
        self.byte_count = oser.UBInt8()
        self.read_register_value = oser.Array(length=lambda ctx: ctx.byte_count.get() // 2, prototype=oser.UBInt16)


class ReadFifoQueueRequest(oser.ByteStruct):
    """
    Representation of a read fifo queue request.
    """

    def __init__(self) -> None:
        super(ReadFifoQueueRequest, self).__init__()
        self.fifo_pointer_address = oser.UBInt16()


class ReadFifoQueueResponse(ExceptionResponse):
    """
    Representation of a read fifo queue response.
    """

    def __init__(self) -> None:
        super(ReadFifoQueueResponse, self).__init__()
        self.byte_count = oser.UBInt8()
        self.fifo_count = oser.UBInt16()
        self.fifo_value_register = oser.Array(length=lambda ctx: ctx.fifo_count.get(), prototype=oser.UBInt16)


class EncapsulatedInterfaceTransportRequest(oser.ByteStruct):
    """
    Representation of an encapsulated interface transport request.
    """

    def __init__(self) -> None:
        super(EncapsulatedInterfaceTransportRequest, self).__init__()
        self.mei_type = oser.UBInt8()
        self.read_device_id_code = oser.If(
            lambda ctx: ctx.mei_type.get() == 0x0E,
            oser.Enum(
                prototype=oser.UBInt8,
                values={
                    "basic": 0x01,
                    "regular": 0x02,
                    "extended": 0x03,
                    "specific": 0x04,
                },
            ),
        )
        self.object_id = oser.If(
            lambda ctx: ctx.mei_type.get() == 0x0E,
            oser.Enum(
                prototype=oser.UBInt8,
                values={
                    "vendor_name": 0x00,
                    "product_code": 0x01,
                    "major_minor_version": 0x02,
                    "vendor_url": 0x03,
                    "product_name": 0x04,
                    "model_name": 0x05,
                    "user_application_name": 0x06,
                },
            ),
        )


class EncapsulatedInterfaceTransportResponse(ExceptionResponse):
    """
    Representation of an encapsulated interface transport response.
    """

    def __init__(self) -> None:
        super(EncapsulatedInterfaceTransportResponse, self).__init__()
        self.mei_type = oser.UBInt8()
        self.read_device_id_code = oser.If(
            lambda ctx: ctx.mei_type.get() == 0x0E,
            oser.Enum(
                prototype=oser.UBInt8,
                values={
                    "basic": 0x01,
                    "regular": 0x02,
                    "extended": 0x03,
                    "specific": 0x04,
                },
            ),
        )
        self.conformity_level = oser.If(lambda ctx: ctx.mei_type.get() == 0x0E, oser.UBInt8())
        self.more_follows = oser.If(lambda ctx: ctx.mei_type.get() == 0x0E, oser.UBInt8())
        self.next_object_id = oser.If(lambda ctx: ctx.mei_type.get() == 0x0E, oser.UBInt8())
        self.number_of_objects = oser.If(lambda ctx: ctx.mei_type.get() == 0x0E, oser.UBInt8())
        self.list_of_objects = oser.If(
            lambda ctx: ctx.mei_type.get() == 0x0E,
            oser.Array(
                length=lambda ctx: ctx.number_of_objects.get(),
                prototype=ReadDeviceIdentificationObject,
            ),
        )


class ReadDeviceIdentificationObject(oser.ByteStruct):
    """
    Representation of read device identification object.
    """

    def __init__(self) -> None:
        super(ReadDeviceIdentificationObject, self).__init__()
        self.object_id = oser.Enum(
            prototype=oser.UBInt8,
            values={
                "vendor_name": 0x00,
                "product_code": 0x01,
                "major_minor_version": 0x02,
                "vendor_url": 0x03,
                "product_name": 0x04,
                "model_name": 0x05,
                "user_application_name": 0x06,
            },
        )
        self.length = oser.UBInt8()
        self.value = oser.Data(length=lambda ctx: ctx.length.get())


class SerialRTUModbusRequest(oser.ByteStruct):
    """
    Modbus Serial RTU request.
    """

    def __init__(self) -> None:
        super(SerialRTUModbusRequest, self).__init__()
        self.address = oser.UBInt8()
        self.pdu = RequestPDU()
        self.crc = ModbusCrc()


class SerialRTUModbusResponse(oser.ByteStruct):
    """
    Modbus Serial RTU response.
    """

    def __init__(self) -> None:
        super(SerialRTUModbusResponse, self).__init__()
        self.address = oser.UBInt8()
        self.pdu = ResponsePDU()
        self.crc = ModbusCrc()


class SerialASCIIModbusRequest(oser.ByteStruct):
    """
    Modbus Serial ASCII request.
    """

    def __init__(self) -> None:
        super(SerialASCIIModbusRequest, self).__init__()
        self.address = oser.UBInt8()
        self.pdu = RequestPDU()
        self.lrc = ModbusLRC()


class SerialASCIIModbusResponse(oser.ByteStruct):
    """
    Modbus Serial ASCII response.
    """

    def __init__(self) -> None:
        super(SerialASCIIModbusResponse, self).__init__()
        self.address = oser.UBInt8()
        self.pdu = ResponsePDU()
        self.lrc = ModbusLRC()


class TCPModbusRequest(oser.ByteStruct):
    """
    Modbus TCP request.
    """

    def __init__(self) -> None:
        super(TCPModbusRequest, self).__init__()
        self.transaction_identifier = oser.UBInt16()
        self.protocol_identifier = oser.UBInt16()
        self.length = oser.UBInt16()
        self.unit_identifier = oser.UBInt8()
        self.pdu = RequestPDU()


class TCPModbusResponse(oser.ByteStruct):
    """
    Modbus TCP response.
    """

    def __init__(self) -> None:
        super(TCPModbusResponse, self).__init__()
        self.transaction_identifier = oser.UBInt16()
        self.protocol_identifier = oser.UBInt16()
        self.length = oser.UBInt16()
        self.unit_identifier = oser.UBInt8()
        self.pdu = ResponsePDU()
